@@ -48,10 +48,16 @@ struct gvm_configuration {
 	gboolean autovideocam;
 	char *autovideocam_command;
 	
-	gboolean autopalmsync;
-	char *autopalmsync_command;
+	gboolean autopilot;
+	char *autopilot_command;
+	
+	gboolean autopocketpc;
+	char *autopocketpc_command;
 	
 	char *eject_command;
+	
+	char *mount_command;
+	char *unmount_command;
 };
 
 /* where our settings are rooted in the gconf tree */
