@@ -61,6 +61,9 @@ static GHashTable *mount_table = NULL;
 /* List of UDI's of all volumes mounted during the lifetime of the program */
 static GSList *mounted_volumes = NULL;
 
+/* Table of active dialogs indexed by the udi of the device in question */
+static GHashTable *dialogs = NULL;
+
 
 typedef enum {
 	TYPE_BOOL,
@@ -93,6 +96,8 @@ enum {
 	AUTORUN,
 	AUTORUN_PATH,
 	EJECT_COMMAND,
+	USER_MOUNT_COMMAND,
+	USER_UNMOUNT_COMMAND
 };
 
 static struct {
@@ -125,6 +130,28 @@ static struct {
 	{ GCONF_ROOT "autorun",                   TYPE_BOOL,   &config.autorun                   },
 	{ GCONF_ROOT "autorun_path",              TYPE_STRING, &config.autorun_path              },
 	{ GCONF_ROOT "eject_command",             TYPE_STRING, &config.eject_command             },
+	{ GCONF_ROOT "mount_command",             TYPE_STRING, &config.mount_command             },
+	{ GCONF_ROOT "unmount_command",           TYPE_STRING, &config.unmount_command           },
+};
+
+/* mount commands to try in order of preference */
+static struct {
+	const char *command;  /* full command-line */
+	char *path;           /* path to mount program - run-time discovered */
+} mount_commands[] = {
+	{ NULL,                         NULL },  /* user override mount command */
+	{ "pmount-hal " PMOUNT_OPTIONS, NULL },  /* pmount-hal */
+	{ "mount " MOUNT_OPTIONS,       NULL },  /* mount */
+};
+
+/* unmount commands to try in order of preference */
+static struct {
+	const char *command;  /* full command-line */
+	char *path;           /* path to mount program - run-time discovered */
+} unmount_commands[] = {
+	{ NULL,                         NULL },  /* user override unmount command */
+	{ "pumount " PUMOUNT_OPTIONS,   NULL },  /* pumount */
+	{ "umount " UMOUNT_OPTIONS,     NULL },  /* umount */
 };
 
 static GHashTable *gvm_settings_hash = NULL;
@@ -340,6 +367,7 @@ prompt_response_cb (GtkWidget *dialog, int response, GvmPromptCtx *ctx)
 			g_object_unref (gconf);
 		}
 		
+		g_hash_table_remove (dialogs, ctx->udi);
 		ctx->callback (ctx, response);
 		gtk_widget_destroy (dialog);
 		gvm_prompt_ctx_free (ctx);
@@ -391,9 +419,10 @@ gvm_prompt (GvmPromptCtx *ctx, int argc, char **argv)
 {
 	GtkWidget *dialog, *hbox, *vbox, *image, *label, *check = NULL;
 	GvmPrompt prompt = ctx->prompt;
-	GnomeIconTheme *theme;
+	GtkIconTheme *theme;
 	GConfClient *gconf;
 	GError *err = NULL;
+	GdkPixbuf *pixbuf;
 	const char *text;
 	int response, i;
 	char *buf;
@@ -449,12 +478,12 @@ gvm_prompt (GvmPromptCtx *ctx, int argc, char **argv)
 	gtk_container_set_border_width ((GtkContainer *) hbox, 12);
 	
 	/* set the icon */
-	theme = gnome_icon_theme_new ();
+	theme = gtk_icon_theme_get_default ();
 	image = gtk_image_new ();
 	
-	if ((buf = gnome_icon_theme_lookup_icon (theme, gvm_prompts[prompt].icon, 48, NULL, NULL))) {
-		gtk_image_set_from_file ((GtkImage *) image, buf);
-		g_free (buf);
+	if ((pixbuf = gtk_icon_theme_load_icon (theme, gvm_prompts[prompt].icon, 48, 0, NULL))) {
+		gtk_image_set_from_pixbuf ((GtkImage *) image, pixbuf);
+		g_object_unref (pixbuf);
 	} else {
 	}
 	
@@ -514,6 +543,8 @@ gvm_prompt (GvmPromptCtx *ctx, int argc, char **argv)
 	
 	g_object_unref (gconf);
 	
+	g_hash_table_insert (dialogs, ctx->udi, dialog);
+	
 	g_signal_connect (dialog, "response", G_CALLBACK (prompt_response_cb), ctx);
 	
 	gtk_widget_show (dialog);
@@ -549,7 +580,7 @@ to_be_or_not_to_be (void)
 static void
 gvm_load_config (void)
 {
-	unsigned int i;
+	size_t i;
 	
 	gvm_settings_hash = g_hash_table_new (g_str_hash, g_str_equal);
 	
@@ -560,6 +591,14 @@ gvm_load_config (void)
 				gconf_client_get_string (config.client, gvm_settings[i].key, NULL);
 			dbg ("setting[%d]: string: %s = %s\n", i, strrchr (gvm_settings[i].key, '/') + 1,
 			     *((char **) gvm_settings[i].var));
+			
+			if (i == USER_MOUNT_COMMAND) {
+				/* set the mount_commands default mount program */
+				mount_commands[0].command = config.mount_command;
+			} else if (i == USER_UNMOUNT_COMMAND) {
+				/* set the unmount_commands default unmount program */
+				unmount_commands[0].command = config.unmount_command;
+			}
 		} else if (gvm_settings[i].type == TYPE_BOOL) {
 			*((int *) gvm_settings[i].var) =
 				gconf_client_get_bool (config.client, gvm_settings[i].key, NULL);
@@ -601,6 +640,18 @@ gvm_config_changed (GConfClient *client __attribute__((__unused__)),
 		*((char **) gvm_settings[which].var) = g_strdup (gconf_value_get_string (value));
 		dbg ("setting changed: string: %s = %s\n", strrchr (gvm_settings[which].key, '/') + 1,
 		     *((char **) gvm_settings[which].var));
+		
+		if (which == USER_MOUNT_COMMAND) {
+			/* reset the mount_commands default mount program */
+			mount_commands[0].command = config.mount_command;
+			g_free (mount_commands[0].path);
+			mount_commands[0].path = NULL;
+		} else if (which == USER_UNMOUNT_COMMAND) {
+			/* reset the unmount_commands default unmount program */
+			unmount_commands[0].command = config.unmount_command;
+			g_free (unmount_commands[0].path);
+			unmount_commands[0].path = NULL;
+		}
 	} else if (gvm_settings[which].type == TYPE_BOOL) {
 		*((int *) gvm_settings[which].var) = gconf_value_get_bool (value);
 		dbg ("setting changed: bool: %s = %d\n", strrchr (gvm_settings[which].key, '/') + 1,
@@ -768,7 +819,7 @@ gvm_check_vcd (const char *udi, const char *device, const char *mount_point)
 	
 	is_vcd = gvm_check_dir (mount_point, "vcd");
 	
-	if (is_vcd && config.autoplay_dvd)
+	if (is_vcd && config.autoplay_vcd)
 		gvm_run_command (config.autoplay_vcd_command, udi, device, mount_point);
 	
 	return is_vcd;
@@ -1067,7 +1118,7 @@ autorun_cb (GvmPromptCtx *ctx, int action)
 static void
 gvm_autorun (const char *udi, const char *device, const char *mount_point)
 {
-	gboolean autobrowse = FALSE;
+	gboolean autobrowse = TRUE;
 	GvmPromptCtx *ctx;
 	
 	if (gvm_check_dvd (udi, device, mount_point))
@@ -1231,7 +1282,43 @@ gvm_device_mounted (const char *udi)
 static gboolean
 gvm_device_mount (const char *udi, const char *device, const char *mount_point, int apply_policy)
 {
+	const char *mount_command = NULL;
 	struct _MountPolicy *policy;
+	char **argv;
+	int argc;
+	size_t i;
+	
+	for (i = 0; i < G_N_ELEMENTS (mount_commands); i++) {
+		if (!mount_commands[i].command || mount_commands[i].command[0] == '\0')
+			continue;
+		
+		if (!mount_commands[i].path) {
+			if (g_shell_parse_argv (mount_commands[i].command, &argc, &argv, NULL)) {
+				if (argv[0][0] != '/') {
+					if (!(mount_commands[i].path = g_find_program_in_path (argv[0]))) {
+						/* program doesn't exist or not in path */
+						mount_commands[i].command = NULL;
+						continue;
+					}
+				} else
+					mount_commands[i].path = g_strdup (argv[0]);
+				g_strfreev (argv);
+			} else {
+				/* command-line is busted */
+				mount_commands[i].command = NULL;
+				continue;
+			}
+		}
+		
+		if (access (mount_commands[i].path, F_OK | X_OK) == 0) {
+			mount_command = mount_commands[i].command;
+			goto mount;
+		}
+	}
+	
+	return FALSE;
+	
+ mount:
 	
 	policy = g_new (struct _MountPolicy, 1);
 	policy->udi = g_strdup (udi);
@@ -1240,7 +1327,7 @@ gvm_device_mount (const char *udi, const char *device, const char *mount_point,
 	g_hash_table_insert (mount_table, policy->udi, policy);
 	
 	dbg ("mounting %s...\n", udi);
-	if (gvm_run_command (MOUNT_COMMAND, udi, device, mount_point))
+	if (gvm_run_command (mount_command, udi, device, mount_point))
 		return TRUE;
 	
 	dbg ("mount failed: %s\n", udi);
@@ -1259,7 +1346,44 @@ gvm_device_mount (const char *udi, const char *device, const char *mount_point,
 static gboolean
 gvm_device_unmount (const char *udi, const char *device, const char *mount_point)
 {
-	return gvm_run_command (UNMOUNT_COMMAND, udi, device, mount_point);
+	const char *unmount_command = NULL;
+	char **argv;
+	int argc;
+	size_t i;
+	
+	for (i = 0; i < G_N_ELEMENTS (unmount_commands); i++) {
+		if (!unmount_commands[i].command || unmount_commands[i].command[0] == '\0')
+			continue;
+		
+		if (!unmount_commands[i].path) {
+			if (g_shell_parse_argv (unmount_commands[i].command, &argc, &argv, NULL)) {
+				if (argv[0][0] != '/') {
+					if (!(unmount_commands[i].path = g_find_program_in_path (argv[0]))) {
+						/* command doesn't exist or not in path */
+						unmount_commands[i].command = NULL;
+						continue;
+					}
+				} else
+					unmount_commands[i].path = g_strdup (argv[0]);
+				g_strfreev (argv);
+			} else {
+				/* command-line is busted */
+				unmount_commands[i].command = NULL;
+				continue;
+			}
+		}
+		
+		if (access (unmount_commands[i].path, F_OK | X_OK) == 0) {
+			unmount_command = unmount_commands[i].command;
+			goto unmount;
+		}
+	}
+	
+	return FALSE;
+	
+ unmount:
+	
+	return gvm_run_command (unmount_command, udi, device, mount_point);
 }
 
 /*
@@ -1322,41 +1446,57 @@ gvm_ask_mixed (const char *udi)
 	libhal_free_string (device);
 }
 
-enum {
+
+typedef enum {
 	WRITER_TYPE_NONE,
 	WRITER_TYPE_CDR,
 	WRITER_TYPE_DVD
+} writer_t;
+
+static struct {
+	const char *disc;
+	const char *drive;
+	writer_t type;
+} burners[] = {
+	{ "cd_r",        "storage.cdrom.cdr",       WRITER_TYPE_CDR },
+	{ "cd_rw",       "storage.cdrom.cdrw",      WRITER_TYPE_CDR },
+	{ "dvd_r",       "storage.cdrom.dvdr",      WRITER_TYPE_DVD },
+	{ "dvd_rw",      "storage.cdrom.dvdrw",     WRITER_TYPE_DVD },
+	{ "dvd_ram",     "storage.cdrom.dvdram",    WRITER_TYPE_DVD },
+	{ "dvd_plus_r",  "storage.cdrom.dvdplusr",  WRITER_TYPE_DVD },
+	{ "dvd_plus_rw", "storage.cdrom.dvdplusrw", WRITER_TYPE_DVD },
 };
 
 /*
- * gvm_device_is_writer - is this device capable of writing CDs/DVDs?
+ * gvm_cdrom_media_is_writable - returns the type of media that can be written
  */
-static int
-gvm_device_is_writer (const char *udi)
+static writer_t
+gvm_cdrom_media_is_writable (const char *udi)
 {
-	static const char *writers[] = {
-		"storage.cdrom.cdr",
-		"storage.cdrom.cdrw",
-		"storage.cdrom.dvdr",
-		"storage.cdrom.dvdram",
-		"storage.cdrom.dvdplusr",
-		"storage.cdrom.dvdplusrw"
-	};
+	writer_t retval = WRITER_TYPE_NONE;
+	char *drive = NULL;
+	char *disc = NULL;
 	size_t i;
 	
-	for (i = 0; i < G_N_ELEMENTS (writers); i++) {
-		if (libhal_device_get_property_bool (hal_ctx, udi, writers[i], NULL)) {
-			const char *type;
+	if (!(disc = libhal_device_get_property_string (hal_ctx, udi, "volume.disc.type", NULL)))
+		return FALSE;
+	
+	for (i = 0; i < G_N_ELEMENTS (burners); i++) {
+		if (!strcmp (burners[i].disc, disc)) {
+			if (!(drive = libhal_device_get_property_string (hal_ctx, udi, "info.parent", NULL)))
+				break;
 			
-			type = strrchr (writers[i], '.') + 1;
-			if (!strncmp (type, "dvd", 3))
-				return WRITER_TYPE_DVD;
-			else
-				return WRITER_TYPE_CDR;
+			if (libhal_device_get_property_bool (hal_ctx, drive, burners[i].drive, NULL))
+				retval = burners[i].type;
+			
+			break;
 		}
 	}
 	
-	return WRITER_TYPE_NONE;
+	libhal_free_string (drive);
+	libhal_free_string (disc);
+	
+	return retval;
 }
 
 static void
@@ -1408,11 +1548,10 @@ gvm_run_cdburner (const char *udi, int type, const char *device, const char *mou
 static void
 gvm_cdrom_policy (const char *udi)
 {
-	char *device = NULL;
-	char *drive_udi = NULL;
 	dbus_bool_t has_audio;
 	dbus_bool_t has_data;
 	dbus_bool_t is_blank;
+	char *device = NULL;
 	DBusError error;
 	int type;
 	
@@ -1420,13 +1559,11 @@ gvm_cdrom_policy (const char *udi)
 	has_audio = libhal_device_get_property_bool (hal_ctx, udi, "volume.disc.has_audio", NULL);
 	has_data = libhal_device_get_property_bool (hal_ctx, udi, "volume.disc.has_data", NULL);
 	is_blank = libhal_device_get_property_bool (hal_ctx, udi, "volume.disc.is_blank", NULL);
-	drive_udi = libhal_device_get_property_string (hal_ctx, udi, "info.parent", NULL);
-	device = libhal_device_get_property_string (hal_ctx, udi, "block.device", &error);
 	
-	if (!device) {
+	if (!(device = libhal_device_get_property_string (hal_ctx, udi, "block.device", &error))) {
 		warn ("cannot get block.device: %s", error.message);
 		dbus_error_free (&error);
-		goto out;
+		return;
 	}
 	
 	if (has_audio && !has_data) {
@@ -1438,16 +1575,13 @@ gvm_cdrom_policy (const char *udi)
 		if (config.automount_media && !gvm_udi_is_subfs_mount (udi))
 			gvm_device_mount (udi, device, NULL, TRUE);
 	} else if (is_blank) {
-		if ((type = gvm_device_is_writer (drive_udi)))
+		if ((type = gvm_cdrom_media_is_writable (udi)))
 			gvm_run_cdburner (udi, type, device, NULL);
 	}
 	
 	/** @todo enforce policy for all the new disc types now supported */
 	
- out:
-	
 	libhal_free_string (device);
-	libhal_free_string (drive_udi);
 }
 
 /*
@@ -1604,7 +1738,7 @@ block_device_added (const char *udi, const char *capability __attribute__((__unu
 		}
 	}
 	
-out:
+ out:
 	
 	if (dbus_error_is_set (&error))
 		dbus_error_free (&error);
@@ -1658,12 +1792,12 @@ input_device_added (const char *udi, const char *capability)
 		if (!strcmp (inputs[i].capability, capability)) {
 			autoexec = *inputs[i].autoexec;
 			command = *inputs[i].command;
-			handled = TRUE;
 			break;
 		}
 	}
 	
 	if (autoexec && command) {
+		handled = TRUE;
 		dbus_error_init (&error);
 		if ((device = libhal_device_get_property_string (hal_ctx, udi, "input.device", &error))) {
 			gvm_run_command (command, udi, device, NULL);
@@ -1712,6 +1846,7 @@ printer_device_added (const char *udi, const char *capability __attribute__((__u
 	return TRUE;
 }
 
+/* Note: this list must be sorted alphabetically for the algorithm in hal_device_added to work */
 static struct {
 	const char *capability;
 	DeviceAddedHandler handler;
@@ -1750,7 +1885,7 @@ hal_device_added (LibHalContext *ctx __attribute__((__unused__)),
 	
 	for (i = 0, j = 0; i < G_N_ELEMENTS (devices) && j < n; i++) {
 		int cmp = -1;
-
+		
 		while (j < n && (cmp = strcmp (capabilities[j], devices[i].capability)) < 0)
 			j++;
 		
@@ -1773,7 +1908,12 @@ static void
 hal_device_removed (LibHalContext *ctx __attribute__((__unused__)), 
 		    const char *udi)
 {
+	GtkDialog *dialog;
+	
 	dbg ("Device removed: %s\n", udi);
+	
+	if ((dialog = g_hash_table_lookup (dialogs, udi)))
+		gtk_dialog_response (dialog, GTK_RESPONSE_CANCEL);
 }
 
 /** Invoked when device in the Global Device List acquires a new capability.
@@ -2194,6 +2334,7 @@ main (int argc, char *argv[])
 	signal (SIGTERM, handle_sigterm);
 	
 	mount_table = g_hash_table_new (g_str_hash, g_str_equal);
+	dialogs = g_hash_table_new (g_str_hash, g_str_equal);
 	
 	mount_all (hal_ctx);
 	
